/*
 * Copyright (c) 2020, Peter Abeles. All Rights Reserved.
 *
 * This file is part of Efficient Java Matrix Library (EJML).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ejml.dense.row.linsol.qr;

import org.ejml.UtilEjml;
import org.ejml.data.DMatrixRMaj;
import org.ejml.dense.row.CommonOps_DDRM;
import org.ejml.dense.row.SpecializedOps_DDRM;
import org.ejml.dense.row.decomposition.TriangularSolver_DDRM;
import org.ejml.dense.row.linsol.LinearSolverAbstract_DDRM;
import org.ejml.interfaces.decomposition.QRDecomposition;

/**
 * <p>
 * A solver for a generic QR decomposition algorithm.  This will in general be a bit slower than the
 * specialized once since the full Q and R matrices need to be extracted.
 * </p>
 * <p>
 * It solve for x by first multiplying b by the transpose of Q then solving for the result.
 * <br>
 * QRx=b<br>
 * Rx=Q^T b<br>
 * </p>
 *
 * @author Peter Abeles
 */
@SuppressWarnings("NullAway.Init")
public class LinearSolverQr_DDRM extends LinearSolverAbstract_DDRM {

    private QRDecomposition<DMatrixRMaj> decomposer;

    protected int maxRows = -1;
    protected int maxCols = -1;

    protected DMatrixRMaj Q;
    protected DMatrixRMaj R;

    private DMatrixRMaj Y, Z;

    /**
     * Creates a linear solver that uses QR decomposition.
     */
    public LinearSolverQr_DDRM( QRDecomposition<DMatrixRMaj> decomposer ) {
        this.decomposer = decomposer;
    }

    /**
     * Changes the size of the matrix it can solve for
     *
     * @param maxRows Maximum number of rows in the matrix it will decompose.
     * @param maxCols Maximum number of columns in the matrix it will decompose.
     */
    public void setMaxSize( int maxRows, int maxCols ) {
        this.maxRows = maxRows;
        this.maxCols = maxCols;

        Q = new DMatrixRMaj(maxRows, maxRows);
        R = new DMatrixRMaj(maxRows, maxCols);

        Y = new DMatrixRMaj(maxRows, 1);
        Z = new DMatrixRMaj(maxRows, 1);
    }

    /**
     * Performs QR decomposition on A
     *
     * @param A not modified.
     */
    @Override
    public boolean setA( DMatrixRMaj A ) {
        if (A.numRows > maxRows || A.numCols > maxCols) {
            setMaxSize(A.numRows, A.numCols);
        }

        _setA(A);
        if (!decomposer.decompose(A))
            return false;

        Q.reshape(numRows, numRows, false);
        R.reshape(numRows, numCols, false);
        decomposer.getQ(Q, false);
        decomposer.getR(R, false);

        return true;
    }

    @Override
    public /**/double quality() {
        return SpecializedOps_DDRM.qualityTriangular(R);
    }

    /**
     * Solves for X using the QR decomposition.
     *
     * @param B A matrix that is n by m.  Not modified.
     * @param X An n by m matrix where the solution is written to.  Modified.
     */
    @Override
    public void solve( DMatrixRMaj B, DMatrixRMaj X ) {
        UtilEjml.checkReshapeSolve(numRows, numCols, B, X);

        int BnumCols = B.numCols;

        Y.reshape(numRows, 1, false);
        Z.reshape(numRows, 1, false);

        // solve each column one by one
        for (int colB = 0; colB < BnumCols; colB++) {

            // make a copy of this column in the vector
            for (int i = 0; i < numRows; i++) {
                Y.data[i] = B.get(i, colB);
            }

            // Solve Qa=b
            // a = Q'b
            CommonOps_DDRM.multTransA(Q, Y, Z);

            // solve for Rx = b using the standard upper triangular solver
            TriangularSolver_DDRM.solveU(R.data, Z.data, numCols);

            // save the results
            for (int i = 0; i < numCols; i++) {
                X.set(i, colB, Z.data[i]);
            }
        }
    }

    @Override
    public boolean modifiesA() {
        return decomposer.inputModified();
    }

    @Override
    public boolean modifiesB() {
        return false;
    }

    @Override
    public QRDecomposition<DMatrixRMaj> getDecomposition() {
        return decomposer;
    }

    public QRDecomposition<DMatrixRMaj> getDecomposer() {
        return decomposer;
    }

    public DMatrixRMaj getQ() {
        return Q;
    }

    public DMatrixRMaj getR() {
        return R;
    }
}